# Copyright (c) 2005 Universit de Lige, Nicolas Mazziotta
# $Id: EasXML.pm 197 2006-03-19 09:33:26Z mzi $
# This program is free software; you can redistribute it and/or
# modify it under the same terms as Perl itself.


package XML::EasXML;

use strict;		

#######################################################
#		 Constructor and object building
#######################################################

use XML::LibXML;
use Carp;

=head1 NAME

XML::EasXML - Handle XML::LibXML objects with ease.

=head1 SYNOPSIS

=over 2

=item use XML::EasXML;

=item my $xml = XML::EasXML->new(file => foo.xml);

=item $xml->print;

=item $xml->elt_id("bar")->delete;

=item # alternately

=item my $xml2 = XML::EasXML->new(xml => $xmlDoc);

=back

=cut

=head1 DESCRIPTION

This module provides easy to use methods to access and manipulate an XML::LibXML::Document. 
The main aim is to provide easy access to ids and a way to create them and maintain a list of them.
This feature is more memory consuming, since it creates a hash mapping ids to nodes.
But it allows a more consistent way to create and handle ids (create_id and provide_ids subroutines cannot 
be used when id support is off).

=head1 CONSTRUCTOR

The XML::EasXML object wraps an XML::LibXML::Document object. 
It is therefore DOM-based. 

=over 2

=item B<new>

The method usually takes only one argument, a hash of options, 
in which is given:

=over 2 

=item - either the xml file (file => "foo.xml") or the XML::LibXML::Document to handle (xml => $xmlDoc);

=item - wether the id support is on or off (id_support => 1|0).

=back

=back 

=cut

# FIXME: the interface is not smart enough: should 
# subclass XML::LibXML::Document instead.

sub new {
  my $proto = shift;
  my $class = ref($proto) || $proto;
  my $o = {
    # you can either load xml from file, from string or even from
    # an already existing XML::LibXML::Document (xml)
    file => undef,
    string => undef,
    xml => undef,
    ids => undef,
    last_id => undef,
    # options
    id_support => 1,
    @_
  };
  bless($o, $class);
  if ($o->{xml}) {}
  # create xml from file 
  elsif ($o->{file}) {
    $o->fromFile($o->{file})
  }
  # create xml from string 
  elsif ($o->{string}) {
    $o->fromFile($o->{file})
  }
  $o->buid_ids_table if $o->{id_support};
  return $o;
}



#######################################################
#		 Properties & options
#		 see below for ids
#######################################################

	#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#
	#		 Properties
	#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

=head1 METHODS

=over 2

=item 1. Properties Access

=over 2

=item B<xml>

parameters: $xml? - $xml is an XML::LibXML::Document.

returns: the value of $o->{xml}.

If $xml is passed, sets $o->{xml} to $xml.

=cut

sub xml {
	my ($o, $xml) = @_;
	return $o->{xml} unless $xml;
	$o->{xml} = $xml;
	return $xml;
}

=item B<file>

parameters: $file? - $file is the name of an existing file

returns: the value of $o->{file}.

If $file is passed, sets $o->{file} to $file.

=cut

sub file {
	my ($o, $xml) = @_;
	return $o->{file} unless $xml;
	$o->{file} = $xml;
	return $xml;
}

	#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#
	#		 Options
	#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

=item B<id_support>

parameters: $boolean

returns: the value of $o->{id_support}.

If $boolean is passed, sets $o->{id_support} to $boolean. 
If id support is turned on this way, the ids table is built; 
if turned off, the table is undefined.

=cut
sub id_support {
	my ($o, $boolean) = @_;
	return $o->{id_support}  unless defined $boolean;
	if ($boolean and ! $o->{id_support}) {
		$o->buid_ids_table;
	} elsif (! $boolean and $o->{id_support}) {
		$o->{ids} = undef;
	}
	return $o->{id_support} = $boolean;
}


#######################################################
#		 XML and file data access
#######################################################

	#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#
	#		 File IO
	#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

=back

=item 2. File IO

=over 2

=item B<fromFile>

parameters: $file

returns: XML::LibXML object.

Builds an XML::LibXML object from $file and sets it as $o->{xml} value.

=cut

sub fromFile {
	my ($o, $file) = @_;
	my $warn = "";
	$warn = 1 unless -f $file;
	if (!$warn) {
		my $parser = XML::LibXML->new();
		my $xml = "";
		if (eval{$xml = $parser->parse_file($file)}) {
			$o->{xml} = $xml;
			return $xml
		} else {
			carp "Parsing error: $@";
			return 0;
		}
	} else {
		carp "$file: no such file!";
		return 0;
	}
}

sub fromString {
  my ($o, $string) = @_;
  my $parser = XML::LibXML->new();
  my $xml = "";
  if (eval{$xml = $parser->parse_string($string)}) {
    $o->{xml} = $xml;
    return $xml
  } else {
    carp "Parsing error: $@";
    return 0;
  }
}

=item B<toFile>

parameters: $file?

returns: true.

Serializes $o->{xml} to the file named $file, or to $o->{file} by default.

=cut

sub toFile {
	my ($o, $file) = @_;
	$file = $o->file unless $file;
	return $o->xml->toFile($file, 1);
}
	

 	#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#
	#		 Ids handling
	#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

=back 

=item 3. Ids Management

=over 2

=item B<buid_ids_table>

parameters: $file

returns: $o->{ids}.

Builds the id table of the current XML::LibXML::Document as a hash where keys are the value of the id attribute and values the reference to the XML::LibXML::Element that has this id. 

=cut

sub buid_ids_table {
	my ($o) = @_;
	foreach ( $o->xml->findnodes("//attribute::*[name() = 'id' or name() = 'Id' or name() = 'ID']") ) {
		$o->{ids}{$_->value} = $_->getOwnerElement;
	}
	return $o->{ids}
}

=item B<create_id>

parameters: $node, $prefix?

returns: the string value of the newly created id.

Provides a brand new Id to the node if it has none. 
The Id has the form of the $prefix followed by a number ($prefix = "_" in case it is not defined).
Cannot be used if id support is off.

=cut

sub create_id {
	my ($o, $node, $prefix) = @_;
	carp "Id support is off! Cannot create id." unless $o->{id_support};
	unless ($node->att("id")) {
		my $n = $o->{last_id};
		$prefix = "_" unless $prefix;
		while (exists $o->{ids}{$prefix.++$n}){}
		my $id = $prefix.$n;
		$o->{last_id} = $n;
		$node->setAttribute("id",$id);
		$o->{ids}{$id} = $node;
		return $id;
	}
	return 0;
}

sub force_id {
	my ($o, $node, $id) = @_;
	$node->setAttribute("id",$id);
	$o->{ids}{$id} = $node;
	return $id;
}


=item B<provide_ids>

parameters: $prefix, $name

returns: true.

Provides a brand new Id to the nodes named $name (all nodes if $name is not defined). 
The ids have the form of the $prefix followed by a number ($prefix = "_" in case it is not defined).
Cannot be used if id support is off.

=cut

sub provide_ids {
	my ($o, $prefix, $name) = @_;
	carp "Id support is off! Cannot create id." unless $o->{id_support};
	$name = "*" unless $name;
	foreach ($o->xml->findnodes("//$name"."[not(\@id)]")) {
		$o->create_id($_,$prefix);
	}
	return 1;
}

=item B<elt_id>

parameters: $id

returns: the XML::LibXML::Element matching the $id

Finds the node that has the id $id. (Faster if the id support is on.)

=cut


sub elt_id {
	my ($o, $id) = @_;
	# faster access when id support is on
	if ($o->{id_support}) { return $o->{ids}{$id} }
	return $o->xml->elt_id($id)
}

sub target_id {
  # retourne l'id auquel renvoie l'lment qui a l'id $id
  my ($o, $id) = @_;
  my $pointer = $o->elt_id($id);
  return "" unless $pointer;
  my $target = "";
  return $target if $target = $pointer->att("target"); 
  return $target if $target = $pointer->att("id-ref"); 
  return "";
}

#######################################################
#		 XML
#######################################################

=back

=item 4. Nodes manipulation

=over 2

=item B<delete>

parameters: the XML::LibXML::Element to delete

returns: the deleted node

=cut


sub delete {
  # If $boolean is true, keep intact all nodes which have
  # an @id-ref pointing to the deleted node. 
  # Don't do it unless you know what you are doing. 
  # By default, delete them.
  my ($o, $node, $boolean) = @_;

  unless ($o->{id_support} or $boolean) { carp("Deleting nodes without id support may lead to severe errors.") }
  my $id = $node->id;
  my %removed_ids = ();
  # removes node from id-list
  if ($o->{id_support}) {
    $removed_ids{$id} = 1 if $id;
    # same for descendants
    foreach ($node->findnodes(".//\@id")) {
      $removed_ids{$_->value} =1 if $_;
    }
  }
  if ($id and ! $boolean) {
    foreach my $removed (keys %removed_ids) {
      foreach ( $node->ownerDocument->findnodes("//*[\@id-ref = '$removed']") ) {
        $o->delete($_)# if $removed_ids{$_->att("id-ref")};
      }
      delete $o->{ids}{$removed};
    }
  }
  # FIXME: does not work with documentNode
  return $node->parentNode->removeChild($node);;
}


#sub delete {
#	my ($o, $node) = @_;
#	my $id = $node->id;
#	# removes node from id-list
#	delete $o->{ids}{$id} if $id;
#	return $node->parentNode->removeChild($node);
#}


# removes a node, but not its children

sub erase {
  my ($o, $node) = @_;
  my %updids = ();
  foreach my $inside ($node->childNodes) {
		next if $inside->nodeType == 3;
    my $clone = $node->removeChild($inside);#->cloneNode(1);
    $node->parentNode->insertBefore($clone, $node);
    foreach ($clone->findnodes("./descendant-or-self::*")) {
      next if $_->nodeType == 3;
      my $id = $_->id;
      $updids{$id} = $_;
    }
  }
  $o->delete($node, 1);
  # la table des ids est mise  jour aprs l'effacement des anciens noeuds
  $o->{ids} = {%{$o->{ids}}, %updids};
}


	#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#
	#		 Gateway for root & document methods
	#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#


	
sub root {
	my ($o) = @_;
	return $o->{xml}->root;
}

=back

=item 5. Serialization

=over 2

=item B<toString>

parameters: string format number as in XML::LibXML::Document function

returns: the serialized $o->{xml}

=cut

sub toString {
	my ($o, $format) = @_;
	$format = "0" unless $format;
	return $o->root->toString($format);
}

=item B<print>

parameters: string format number as in XML::LibXML::Document function

returns: true 

Performs a print on the result of $o->toString($format).

=cut

sub print {
	my ($o, $format) = @_;
	$format = "0" unless $format;
	print $o->root->toString($format) || return 0;
}


		
	#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#
	#		 true XML methods
	#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#
	
=back

=back

=head1 XML::LibXML EXTENSION

=over 2

=item 1. XML::LibXML::Document Extension

=over 2

=item B<elt_id>

parameters: $id

returns: XML::LibXML::Element that has this id.

=cut 

sub XML::LibXML::Document::elt_id {
	my ($document, $id) = @_;
	foreach ($document->findnodes("//*[\@id = '$id'][1]")) {return $_}
	return 0;
}


=item B<root>

parameters: none.

returns: the document element.

=cut 

sub XML::LibXML::Document::root {
	my ($document) = @_;
	return $document->documentElement;
}

=back

=item 2. XML::LibXML::Node Extension

=over 2

=item B<print>

parameters: string format number as in XML::LibXML::Node function.

returns: true.

Performs a print on the result of $node->toString($format).

=cut

sub XML::LibXML::Node::print {
	my ($node, $format) = @_;
	$format = 0 unless $format;
	print $node->toString($format);
	return 1;
}

=item B<root>

parameters: none.

returns: the document element owning the node.

=cut

sub XML::LibXML::Node::root {
	my ($node) = @_;
	return $node->ownerDocument->root;
}

=back

=item 3. XML::LibXML::Element Extension

=over 2

=cut 

=item B<att>

parameters: $name, $value?

returns: the value of the attribute $name owned by the element.

Changes the value of the attribute $name to $value if $value is defined.

=cut


sub XML::LibXML::Element::att {
	my ($node, $name, $value) = @_;
	return 0 unless $name;
	if ($name =~ /^id$/ and $value) {
		carp "It is better to use 'create_id' method to handle id attributes! Ids table might be broken.";
	}
	return $node->getAttribute($name) unless defined $value;
	$node->setAttribute($name, $value);
	return $value;
}


=item B<id>

parameters: none.

returns: the value of the id owned by the element.

=cut

sub XML::LibXML::Element::id {
  my ($node) = @_;
  my $id = "";
  if (
    $id = $node->getAttribute("id")
      or
    $id = $node->getAttribute("Id")
      or
    $id = $node->getAttribute("ID")
  ) {return $id}
  else {
    return 0;
  }
}

sub XML::LibXML::Element::trimmed_text {
  my ($node) = @_;
  my $text = $node->textContent;
  $text =~ s/^\s*(.*?)\s*$/$1/;
  $text =~ s/\s+/ /sg;
  return $text;
}


sub XML::LibXML::Element::append {
  my ($o, $elt) = @_;
  return $o->appendChild($elt);
}

sub wrap {
  # emballe l'lment $elt dans un nouvel lment au nom 
  # TODO: s'arranger pour qu'on puisse wrapper des arrays d'elts
  # $newname et aux attributs $atts = {cl => valeur}
  my ($o, $elt, $newname => $atts) = @_;
  $atts = {} unless $atts;  
  my $newelt = XML::LibXML::Element->new("$newname");
  
  map {  
    unless ($_ =~ /^id$/i) {
      $newelt->att($_, $atts->{$_})
    } else {
      $atts->{id} = $atts->{$_};
    }
  } keys %$atts;
  
  my $parent = $elt->parentNode;
  $parent->insertBefore($newelt, $elt);
  
  my $newid = $newelt->id;
  my $oldid = $elt->id;
  my $wrapped = $elt->cloneNode(1);
  
  $parent->removeChild($elt);
  $newelt->appendChild($wrapped);
  
  if ($oldid) { $o->{ids}{$oldid} = $wrapped; }
  if (my $not_safe = $atts->{id}) {
    my $id = $o->create_id($newelt) ;
    warn("I created a safe id ('$id') instead of the one proposed ('$not_safe').")
  };

  return $newelt;
}

sub XML::LibXML::Element::concat {
  my ($o, $elt) = @_;
  eval { return $o->addSibling($elt); }
}

sub  XML::LibXML::Element::prev {
  my ($o, $name) = @_;
  $name = "*" unless $name;
  map { return $_ } reverse $o->findnodes(".//preceding::$name"."[1]");
  return 0;
}

sub  XML::LibXML::Element::prevSibling {
	my ($o, $name) = @_;
	$name = "*" unless $name;
	map { return $_ } reverse $o->findnodes("./preceding-sibling::$name"."[1]");
	return 0;
}

sub  XML::LibXML::Element::next {
	my ($o, $name) = @_;
	$name = "*" unless $name;
	map { return $_ } $o->findnodes(".//following::$name"."[1]");
	return 0;
}

sub  XML::LibXML::Element::nextChild {
	my ($o, $name) = @_;
	$name = "*" unless $name;
	map { return $_ } $o->findnodes("./$name"."[1]");
	return 0;
}



sub  XML::LibXML::Element::nextSibling {
	my ($o, $name) = @_;
	$name = "*" unless $name;
	map { return $_ } $o->findnodes("./following-sibling::$name"."[1]");
	return 0;
}

1;
__END__

=head1 CAVEATS

B<Id Support.> If you rely on id support, be sure of what you do if you  
remove nodes via LibXML methods rather than the ones provided here or create ids "by hand".

=head1 TODO

=head1 BUGS

Please report any bug to the author.

=head1 SEE ALSO

L<XML::LibXML>.

=head1 AUTHOR

Nicolas Mazziotta <NicolasMazziotta@ulgacbe>

=head1 COPYRIGHT

2005 Universit de Lige (Belgium), Nicolas Mazziotta.
This program is free software; you can redistribute it and/or
modify it under the same terms as Perl itself.
