# Copyright (c) 2005 Universit de Lige, Nicolas Mazziotta
# $Id: Khartes.pm 197 2006-03-19 09:33:26Z mzi $
# This program is free software; you can redistribute it and/or
# modify it under the same terms as Perl itself.

use strict;		

#######################################################
package XML::Khartes;
#######################################################

=head1 NAME

XML::Khartes - An XML::EasXML package to handle Kharts files

=head1 SYNOPSIS

=over 2

=item use XML::Khartes

=item my $k = XML::Khartes->new(file => "foo.kml");

=item my $el = $k->elt_id("_n");

=item $k->delete($el);

=back

=cut


require XML::EasXML;
require XML::Khartes::Lexicon; # TODO: enlever ceci ds que possible
require XML::Khartes::Syntax; 
require XML::Khartes::Analysis; 
require XML::Khartes::Text;
require XML::Khartes::IO;

use Carp;

our @ISA = ("XML::EasXML");
our $AUTOLOAD;

=head1 CONSTRUCTOR

=over 2

=item B<new>

=item The method usually takes only one argument, a hash of options, 
in which is given:

=over 2 

=item - either the xml file (file => "foo.xml") or the XML::LibXML::Document to handle (xml => $xmlDoc);

=item - wether the file should be opened read-only or not (new(isLocked => 1) opens it read-only; see $o->isLocked method).

=back

=back 

=cut

sub new {
  my $class = shift;
  my $self  = $class->SUPER::new(@_);
  # Adds a new property set when file is read or written. 
  # It says whether file is locked or not
  $self->{"lock"} = 0;
  # Autoload functions
  $self->{_auto} = {
    # General information
    year => "info",
    month => "info",
    day => "info",
    hand => "info",
    type => "type",
    dimension => "info",
    institution => "info",
    archive => "info",
    v => "info",
  };
  bless ($self, $class);        
  $self->{"lock"} = $self->isLocked unless $self->{"lock"};
  return $self;
}

#=====================================================#
sub  word_pos
#=====================================================#
{
	my ($o, $id) = @_;
	unless ($o->{_words_pos}) {
		my $n = 0;
		foreach ($o->xml->findnodes("/charte/texte/recto//w")) {
			$o->{_words_pos}{$_->id} = ++$n
		}
	}
	return $o->{_words_pos}{$id}
}

#=====================================================#
sub  first_word
#=====================================================#
{
	my ($o) = @_;
	return $o->{__first_word} if $o->{__first_word};
	my $w = $o->xml->root->next("w");
	while ($w->trimmed_text =~ /^\s*$/) {
		$w = $w->next("w");
	}
	$o->{__first_word} = $w;
	return $w;
}

#=====================================================#
sub  last_word
#=====================================================#
{
	my ($o) = @_;
	return $o->{__last_word} if $o->{__last_word};
	foreach my $w ($o->xml->findnodes("/charte/texte/recto//w[last()]")) {
		while ($w->trimmed_text =~ /^\s*$/) {
			$w = $w->prev("w");
		}
		$o->{__last_word} = $w;
		return $w;
	}
}

#=====================================================#
sub word_count
#=====================================================#
{
	my ($o) = @_;
	return scalar keys %{$o->{_words_pos}} if $o->{_words_pos};
	my $n = 0;
	foreach ($o->xml->findnodes("/charte/texte/recto//w")) {++$n}
	return $n;
}

#=====================================================#
sub ponct_count
#=====================================================#
{
	my ($o) = @_;
	my $n = 0;
	foreach ($o->xml->findnodes("/charte/texte/recto//ponct")) {++$n}
	return $n;
}


#=====================================================#
sub line_count 
#=====================================================#
{
	my ($o) = @_;
	return $o->{__line_count} if $o->{__line_count};
	my $n = 0;
	foreach ($o->xml->findnodes("/charte/texte/recto//l")) {++$n}
	$o->{__line_count} = $n;
	return $n;
}

=head1 METHODS

=over 2

=item 1. Properties Access

=over 2

=item B<corpus>

=item I<parameters>: none

=item I<returns>: the owner corpus.

=item Read-only

=cut

# FIXME: un texte devrait pouvoir appartenir  plusieurs corpus
sub corpus {
  my ($o) = @_;
  return $o->{corpus};
}

=item B<info>

=item I<parameters>: $key, $value?, where $key = archive|day|dimension|hand|institution|month|type|year

=item I<returns>: the value of the $key property.

=item Sets the $o->{key} property to $value if the latter is defined. The object can AUTOLOAD methods named after possible $key.

=cut

sub info {
  my ($o, $key, $value) = @_;
  my %keys = (
    archive => "archive",
    day => "jour",
    dimension => "dimensions",
    hand => "main",
    institution => "fonds",
    month => "mois",
    type => "type",
    year => "an",
    v => "version",
  );
  $key = $keys{$key};
  if ($value) { 
    return $o->root->att($key, $value);
  } else {
    return $o->root->att($key);
  }
}

sub time {
  my ($o) = @_;
  return substr($o->year,0,4) . "-" . $o->month . "-" . $o->day;
}

=item B<lexicon>

=item I<parameters>: none.

=item I<returns>: the lexicon XML::EasXML document of the corpus owning the document.

=cut

sub lexicon {
  my ($o) = @_;
  return $o->corpus->lexicon;
}

=item B<reference>

=item I<parameters>: none.

=item I<returns>: a string containing a concatenation the "year" and "institution" properties

=cut

sub reference {
  my ($o) = @_;
  return $o->year . $o->institution;
}

#=====================================================#
 sub  bibtex
#=====================================================#
{
	my ($o) = @_;
	return $o->xml->root->att("bibtex");
}


=item B<isLocked>

=item I<parameters>: none.

=item I<returns>: the value of $o->{isLocked} (0, 1 or the id of the locker, see $o->lock method).

=cut

sub isLocked {
  my ($o) = @_;
  return $o->root->att("lock");
}

=item B<lock>

=item I<parameters>: $name?.

=item I<returns>: 1 or $name if defined.

=item Locks the file for writing. Sets $o->{isLocked} to $name (or 1 if $name is not defined).


=cut

sub lock {
  my ($o, $lock) = @_;
  $lock = 1 unless $lock; # allows to lock w/ uid
  $o->root->att("lock",$lock);
  return $lock;
}

=item B<unlock>

=item I<parameters>: none.

=item I<returns>: 1.

=item Unlocks the file for writing.

=cut

sub unlock {
  my ($o) = @_;
  $o->root->att("lock","0");
  return 1;
}

=back 

=item 2. File IO

=over 2

=item B<toFile>

=item Same as XML::EasXML->toFile, but checks wether the file is locked or not and refuses to write it if true.

=cut

sub toFile {
  my ($o, $file) = @_;
  $file = $o->file() unless $file;
  if ($o->isLocked) { if ($o->{"lock"}) { 
    carp "$file is locked! I cannot write in it. Unlock it first.";
    return 0 ; 
  } }
  $o->{"lock"} = 0;
  $o->xml->toFile($file, 1);
}

=back

=item 3. Ids Management

=over 2

=item B<check_refs>

=item I<parameters>: none.

=item I<returns>: true if cross references are right, false on any error.

=cut


sub check_refs {
  my ($o) = @_;
  my $return = 1;
  foreach ($o->xml->findnodes("//\@id-ref")) {
    my $id = $_->value;
    unless ($o->elt_id($id)) {
      carp("$id is targeted, but does not exist!");
      $return = 0;
    }
  }
  # todo: return a list of elts that target the missing id
  return $return;
}


sub part {
  # FIXME: it does not work
  my ($o, $part) = @_;
  my %parts = (
    text => "texte",
    syntax => "syntaxe",
    diplo => "diplo",
    analysis => "analyse"
  );
  my $partname = $parts{$part};

  foreach ($o->xml->findnodes("/*/$partname"."[1]")) {
    return $_;
  }
}


sub AUTOLOAD {
  my ($o, @args) = shift;
  my $name = $AUTOLOAD;
  $name =~ s/.*:://;
  if (defined $o->{_auto}{$name}) {
    my $function = $o->{_auto}{$name};
    $o->$function($name, @_);
  } else {
    carp("Failed to load AUTOLOAD $name: function name not allowed!")
  }
}

sub DESTROY {
  my $o = shift;
  delete $o->{ids} ;
  delete $o->{xml};
  undef $o;
}


#######################################################
package XML::Khartes::Corpus;
#######################################################

use File::Spec;

# Ce package reprsente une collection de XML::Khartes

sub new {
  my $proto = shift;
  my $class = ref($proto) || $proto;
  my $o = {
    files => [],
    khartes => {},	    # l'ensemble des documents, avec pour chaque une rfrence unique  
    lexicon_file => undef,
    lexicon => undef,
    @_
  };
  bless($o, $class);
  # create xml from file 
  if ($o->{files}) {
    map {$o->add($_)} @{$o->{files}}
  }
  # create lexicon
  if (my $lexicon_file = $o->{lexicon_file}) {
    $o->{lexicon} = $o->_build_lexicon($lexicon_file);
  }
  return $o;
}

sub add {
  # cette mthode prend une liste de noms de fichiers et de rfrences
  #  des objets XML::Khartes (on peut mlanger les deux) et les ajoute au corpus
  my ($o, @add) = @_;
  foreach my $add (@add) {
    if (ref $add eq "XML::Khartes") {
      return $o->_add_object($add);
    } 	
    elsif (-f $add) {
      return $o->_add_file($add);
    } else {
      return 0;
    }
  }
}

sub remove {
  # retire un fichier du corpus d'aprs sa rfrence
  # FIXME: enlever le _file_ de la liste
  my ($o, $ref) = @_;
  my $kh = $o->{khartes}{$ref};
  $kh->DESTROY if $kh;
  delete $o->{khartes}{$ref};
}

sub empty {
  # vide le corpus (retire tous les fichiers
  my ($o) = @_;
  delete $o->{khartes};
  delete $o->{khartes_files};
  delete $o->{files};
  return $o;
}

sub _add_object {
  my ($o, $k) = @_;
  my $ref = $k->reference;
  # TODO: une proprit dans l'objet qui renvoie au lexicon.
  $k->{corpus} = $o;
  $o->{khartes}{$ref} = $k;
  return $k;
}

sub _add_file {
  my ($o, $file) = @_;
  $file = File::Spec->rel2abs( $file );
  my $k = XML::Khartes->new(file => $file); 
  $o->_add_object($k);
  $o->{khartes_files}{$file} = $k;
  return $k;
}

sub _build_lexicon {
  my ($o, $lexicon_file) = @_;
  return $o->{lexicon} = XML::Khartes::Lexicon->new(file => $lexicon_file);
}

sub lexicon {
  my ($o) = @_;
  return $o->{lexicon};
}

sub get {
  my ($o, $ref) = @_;
  return $o->{khartes}{$ref};
}

sub get_file {
  # retourne l'objet XML::Khartes qui correspond au fichier nomm
   my ($o, $file_name) = @_;
   $file_name = File::Spec->rel2abs( $file_name );
   return $o->{khartes_files}{$file_name};
}

sub texts {
  my ($o)  = @_;
  return %{$o->{khartes}};
}

sub toFile {
  my ($o) = @_;
  my %texts = $o->texts;
  foreach (values %texts) {
    $_->toFile;
  }
}
1;
__END__

=back

=head1 Sub Package

For further information, see documentation of the following packages:

=over 2

=item B<XML::Khartes::Syntax>

=item B<XML::Khartes::Export>

=back

=head1 TODO

Document XML::Khartes::Corpus

=head1 BUGS

Please report any bug to the author.

=head1 SEE ALSO

L<XML::EasXML>, L<XML::LibXML>.

=head1 AUTHOR

Nicolas Mazziotta <NicolasMazziotta@ulgacbe>

=head1 COPYRIGHT

2005 Universit de Lige (Belgium), Nicolas Mazziotta.
This program is free software; you can redistribute it and/or
modify it under the same terms as Perl itself.

# vim:sw=2 ts=2
